/* aio-poll-pipe.c
 *	tests async poll functionality on pipes.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/poll.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <asm/unistd.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib_io.h>

int poll_may_complete;
io_context_t ctx;

int done;

void pexit(char *msg)
{
	perror(msg);
	exit(1);
}

void dumpbuf(unsigned char *buffer, int size)
{
	static unsigned char hexit[] = "0123456789abcdef";
	int x=17;

	while (--size >= 0)  {
		if (!--x) {
			x = 16;
			putchar('\n');
		}

		putchar(hexit[*buffer >> 4]);
		putchar(hexit[*buffer & 0xf]);

		buffer++;
	}
	putchar('\n');
}

static void my_done(io_context_t ctx, struct iocb *iocb, long res, long res2)
{

	printf("res: %ld res2: %ld\n", res, res2);
	if (!poll_may_complete) {
		printf("poll completed early!\n");
		exit(1);
	}
	done = 1;
}

int main(int argc, char *argv[])
{
#define BUFSIZE	(4 * 1024)
	struct iocb	*iocb = calloc(1,sizeof(struct iocb));
	struct iocb	*iocbs[] = { iocb };
	char *buffer = malloc(BUFSIZE + 4096);
	long res;
	int pipefds[2];

	if (argc != 1) {
		fprintf(stderr, "usage: aio-poll-pipe\n");
		return 2;
	}

	if ((long)buffer & 4095)
		buffer += 4096 - ((long)buffer & 4095);

	if (pipe(pipefds)) {
		perror("pipe");
		return 1;
	}

	res = io_queue_init(1024, &ctx);
	if (ctx < 0)
		pexit("io_queue_init");

	io_poll(ctx, iocb, my_done, pipefds[1], POLLOUT);
	poll_may_complete = 1;
	io_queue_run(ctx);

	if (!done) {
		printf("Failed!\n");
		return 1;
	}

	done = 0;
	poll_may_complete = 0;
	io_prep_poll(iocb, pipefds[0], POLLIN);
	io_set_callback(iocb, my_done);

	res = io_submit(ctx, 1, iocbs);
	if (res != 1) {
		fprintf(stderr, "io_submit: %ld (%s)\n", res, (res < 0 ? strerror(-res) : ""));
		exit(1);
	}

	io_queue_run(ctx);	/* Just in case the poll triggered early */

	if (3 != write(pipefds[1], "foo", 3)) {
		perror("write");
		return 2;
	}

	poll_may_complete = 1;

	do {
		fprintf(stderr, "calling io_queue_run.\n");
		res = io_queue_run(ctx);
		fprintf(stderr, "io_queue_run: %ld\n", res);
	} while (!done);

	printf("Ok!\n");
	return 0;
}


