/* aio-test.c
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <asm/unistd.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib_io.h>

#define linux_iocb	iocb
int raw_fd = -1;
io_context_t ctx;

int done;

void pexit(char *msg)
{
	perror(msg);
	exit(1);
}

void dumpbuf(unsigned char *buffer, int size)
{
	static unsigned char hexit[] = "0123456789abcdef";
	int x=17;

	while (--size >= 0)  {
		if (!--x) {
			x = 16;
			putchar('\n');
		}

		putchar(hexit[*buffer >> 4]);
		putchar(hexit[*buffer & 0xf]);

		buffer++;
	}
	putchar('\n');
}

static void my_done(io_context_t ctx, struct iocb *iocb, long res, long res2)
{

	printf("res: %ld res2: %ld\n", res, res2);
	done = 1;
}

int main(int argc, char *argv[])
{
#define BUFSIZE	(4 * 1024)
	struct iocb	*iocb = calloc(1,sizeof(struct linux_iocb));
	struct iocb	*iocbs[] = { iocb };
	char *buffer = malloc(BUFSIZE + 4096);
	long res;

	if (argc != 2) {
		fprintf(stderr, "usage: aio-test <file>\n");
		return 2;
	}

	if ((long)buffer & 4095)
		buffer += 4096 - ((long)buffer & 4095);

	dumpbuf(buffer, 512);
	res = io_queue_init(1024, &ctx);
	if (ctx < 0)
		pexit("io_queue_init");

	raw_fd = open(argv[1], O_RDONLY);
	if (raw_fd == -1)
		pexit(argv[1]);

	io_prep_pread(iocb, raw_fd, buffer, BUFSIZE, 0);
	io_set_callback(iocb, my_done);

	res = io_submit(ctx, 1, iocbs);
	if (res != 1) {
		fprintf(stderr, "io_submit: %ld (%s)\n", res, (res < 0 ? strerror(-res) : ""));
		exit(1);
	}

	do {
		fprintf(stderr, "calling io_queue_run.\n");
		res = io_queue_run(ctx);
		fprintf(stderr, "io_queue_run: %ld\n", res);
	} while (!done);

	dumpbuf(buffer, 512);

	return 0;
}


