/* Copyright 2000 Red Hat, Inc.  All Rights Reserved.
 * Written by Benjamin LaHaise <bcrl@redhat.com>
 * This file may be distributed under the terms of the
 * GNU Public License, but is provided with NO WARRANTY.
 */
#define _GNU_SOURCE
#define _FILE_OFFSET_BITS	64
#define NDEBUG

#include <assert.h>
#include <aio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <unistd.h>

#define BUFSIZE		(8 * 1024 * 1024)
#define WRITESIZE	(128 * 1024)
#define READSIZE	(68 * 1024)
#define MINREADSPACE	(512 * 1024)
#define PAGE_SIZE 4096

//char foobuffer[BUFSIZE + PAGE_SIZE];
char *buffer;

int main (int argc, char *argv[])
{
	unsigned long buf_size = BUFSIZE;
	int buf_log = 12;
	struct timeval start, end;
	struct aiocb read_iocb;
	struct aiocb *aio_list[2];
	off_t roff = 0;
	int saw_eof = 0;
	int in_fd = 0;
	int i, k;
	ssize_t ret;
	int count = 1;
	int aio = 1;

	//buffer = foobuffer + PAGE_SIZE - ((unsigned long)foobuffer & (PAGE_SIZE - 1));

	switch (argc) {
	case 5:
		aio = atoi(argv[4]);
	case 4:
		count = atoi(argv[3]);
	case 3:
		buf_log = atol(argv[2]);
	case 2:
		in_fd = open(argv[1], O_RDONLY);
		if (in_fd < 0) {
			perror("open() input");
			return 1;
		}
		break;
	case 1:
	default:
		fprintf(stderr, "usage: foo <filename> <nr_pages> <count> <aio>\n");
		exit(1);
	}

	buf_size = 1UL << buf_log;

	buffer = mmap(NULL, buf_size + 8192, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANON, -1, 0);
	if (buffer == (char *)-1) {
		perror("mmap");
		return 1;
	}

	memset(buffer, 0, buf_size);
	memset(&read_iocb, 0, sizeof(read_iocb));

	for (k=0; k<count; k++) {
		roff = 0;
		saw_eof = 0;

	gettimeofday(&start, NULL);

	do {
		read_iocb.aio_sigevent.sigev_notify = SIGEV_NONE;
		read_iocb.aio_offset = roff;
		read_iocb.aio_fildes = in_fd;
		read_iocb.aio_buf = buffer;
		read_iocb.aio_nbytes = buf_size;

		do {
			switch (aio) {
			//case 2:
			//	ret = linux_aio_read(in_fd, buffer, buf_size, roff);
			//	break;
			case 1:
				ret = aio_read(&read_iocb);
				break;
			case 0:
				ret = pread(in_fd, buffer, buf_size, roff);
				break;
			default:
				ret = -EIO;
			}
		} while (ret < 0 && errno == EINTR);

		if (ret < 0)
			goto read_err;

		i = 0;

		if (1 == aio) {
			aio_list[i++] = &read_iocb;
			do {
				ret = aio_suspend(&aio_list, i, NULL);
				if (ret)
					goto err;

			} while  ((ret = aio_error(&read_iocb)) == EINPROGRESS);
			if (ret < 0) {
				fprintf(stderr, "aio_error\n");
				goto read_err;
			}

			ret = aio_return(&read_iocb);
			if (ret < 0) {
				fprintf(stderr, "aio_return\n");
				goto read_err;
			}
		}

		roff += ret;
		if (!ret)
			saw_eof = 1;
	} while (!saw_eof);

	gettimeofday(&end, NULL);

	fprintf(stderr, "Read %Ld bytes in %Ld us with %lu byte buffers via %s log %d\n", (long long)roff,
		(long long)(end.tv_sec - start.tv_sec) * 1000000 +
		(end.tv_usec - start.tv_usec),
		buf_size,
		aio == 2 ? "linux_aio" : aio ? "aio_read" : "pread",
		buf_log
	);
	}

	return 0;
err:
	perror("aio_suspend");
	fprintf(stderr, "ret: %d\n", ret);

	return 1;

read_err:
	perror("aio_read");
	fprintf(stderr, "ret: %d\n", ret);
	return 1;
}

